#include "kt_serial.h"
#include "kt_hexparser.h"
#include<pthread.h>
#include<sys/types.h>
#include<sys/time.h>
#include<stdio.h>
#include<stdlib.h> 

/* Acknowledge message for the client */
#define bl_ack_ack     'z'
#define bl_ack_nack    'n'
#define bl_ack_err     'e'
#define bl_ack_timeout 't'

int bl_send_char(char data)
{
  return kt_send_char(data);
}

char bl_calc_crc(char * data, unsigned int length)
{
  unsigned int i;
  char crc=0;

#if 0
  for(i=0; i<length;i++)
    printf("%c.",data[i]);
  printf("\r\n");
#endif

  for(i=0; i<length;i++)
    crc+=data[i];

  return crc;
}
int bl_prepare_frame(char * frame,int code, int datacount, int addr, char * data)
{
  int i;

  switch(code)
  {
    case 0:
      frame[0] = code;
      frame[1] = datacount;
      frame[2] = (addr >> 8) & 0x00FF;
      frame[3] = addr & 0x00FF;

      for(i=0; i<datacount;i++)
	frame[4+i] = data[i];

      return (4+datacount);
      break;

    case 1:
      frame[0] = code;
      frame[1] = datacount;
      return 2;
      break;
      
    case 4:
      frame[0] = code;
      frame[1] = data[0];
      frame[2] = data[1];

      return 3;
      break;

    default:
      return 0;
      break;
  }
}

int bl_send_frame(char * data,int datacount)
{
  int i;

  /* Send the frame length */
  bl_send_char(datacount+2);	

  /* Send a data frame */
  kt_send_str(data,datacount);
  //for(i=0; i<datacount; i++)
    //bl_send_char(data[i]);

  /* Sending the crc */
  bl_send_char(bl_calc_crc(data,datacount));

  /* Send the endchar */
  bl_send_char('\n');
}

char bl_read_ack_tmp_fixme = 0;

char bl_get_ack()
{
  struct timeval lastcommand, now, diff;
  struct timezone here;
  char   ack = 0;

  gettimeofday(&lastcommand,&here);

  while(!ack)
  {
    gettimeofday(&now,&here);
    timersub(&now,&lastcommand,&diff);

    /* Timeout */
    if(diff.tv_sec >= 10)
      return bl_ack_timeout;
    /* Read a char from the target */
    /* This call must be non blocking */
    ack = bl_read_ack_tmp_fixme;
  }

  bl_read_ack_tmp_fixme = 0;
  
  return ack;
}

void * task_echo_task(void * arg)
{
  char ans[256];

  while(1)
  {
    kt_read_line(ans,256);
    printf("%s",ans);
    if(!strcmp(ans,"Ack\n"))
      bl_read_ack_tmp_fixme = bl_ack_ack;
    if(!strcmp(ans,"Nck\n"))
      bl_read_ack_tmp_fixme = bl_ack_nack;
    if(!strcmp(ans,"Err\n"))
      bl_read_ack_tmp_fixme = bl_ack_err;
  }
}

int main(int argc, char * argv[])
{
  int i,rc;
  char ack;

  char bl_current_frame[32];
  int  bl_current_framelen;

  enum {bl_state_end, bl_state_waitack, bl_state_data, bl_state_resend, bl_state_restart} bl_client_state;

  pthread_t echo_task;

  FILE* progfile;

  int bl_frame_code;
  int bl_frame_addr;
  int bl_frame_count;
  char bl_frame_data[16];

  unsigned nack_count = 0;

  printf("K-Team uC Programmer v1.0\r\n");

  /* Test args */
  if(argc < 2)
  {
    printf("usage: ucClient <serial device> <hex file>\r\n");
    exit(0);
  }
  if(!argv[1] || !argv[2])
  {
    printf("Error processsing args - null string\r\n");
    exit(0);
  }

  /* Open Serial port */
  rc = kt_open_port(argv[1]);
  if(rc<0)
  {
    fprintf(stderr,"Cannot open serial port (%s) aborting\r\n", argv[1]);
    exit(0);
  }
  kt_set_serial(115200);
  
  /* Open program file */
  progfile = fopen(argv[2],"r");
  if(!progfile)
  {
    fprintf(stderr,"Cannot find %s aborting\r\n", argv[2]);
    exit(0);
  }

  /* Check hex file content */
  if(hexp_valid_hex(progfile)<0)
  {
    fprintf(stderr,"Invalid Hex File aborting\r\n");
    exit(0);
  }
  
  rc = pthread_create(&echo_task, NULL, task_echo_task, NULL);
  if (rc){
    printf("ERROR; return code from pthread_create() is %d\n", rc);
    exit(-1);
  }

  /* Starting the OS loader mode */
  bl_send_char('U');
  bl_send_char('\n');
  /* Wait for the target answer */
  ack = bl_get_ack();
  switch(ack)
  {
    case bl_ack_ack:
      /* ACK received, start prog*/
      printf("Initial Ack, Starting...\r\n");
      break;
    default:
      /* abort */
      printf("Intitial Ack missing, Abort\r\n");
      bl_client_state = bl_state_end;
      break;
  }

  hexp_start(progfile);
  bl_client_state  = bl_state_data;

  while(bl_client_state != bl_state_end)
  {
    switch(bl_client_state)
    {
      case bl_state_data :
	rc = hexp_next_line(progfile,&bl_frame_code,&bl_frame_count,&bl_frame_addr,bl_frame_data);
	if(rc == 1)
	{
	  /* Format frame to send */
	  bl_current_framelen = bl_prepare_frame(bl_current_frame,bl_frame_code,bl_frame_count,bl_frame_addr,bl_frame_data);
	  printf("sending: code:%d length:%d addr:0x%x   \r",bl_frame_code,bl_frame_count,bl_frame_addr);
	  /* Send a frame */
	  bl_send_frame(bl_current_frame,bl_current_framelen);
	  bl_client_state = bl_state_waitack;
	}
	else
	if(rc == 2)  
	  bl_client_state = bl_state_data;
	else
	  bl_client_state = bl_state_end;
	break;

      case bl_state_waitack :
	/* Wait for the target answer */
	ack = bl_get_ack();
	switch(ack)
	{
	  case bl_ack_ack:
	    /* ACK received, next frame */
	    //printf("ack (%u)\r\n",nack_count);
	    bl_client_state = bl_state_data;
	    break;
	  case bl_ack_nack:
	    /* NACK received, resend frame */
	    nack_count++;
	    //printf("nack (%u)\r\n",nack_count);
	    bl_client_state = bl_state_resend;
	    break;
	  case bl_ack_err:
	    /* ERR received, giveup */
	    printf("\nerr (%u)\r\n",nack_count);
	    bl_client_state = bl_state_end;
	    break;
	  case bl_ack_timeout:
	    /* Timeout waiting target answer, giveup */
	    printf("\ntimeout receiving\r\n");
	    bl_client_state = bl_state_end;
	    break;
	  default:
	    /* Unknown answer, ask the target to restart programming */
	    //printf("unknown\r\n");
	    bl_client_state = bl_state_restart;
	    break;
	}
	break;

      case bl_state_resend :
	/* Resend the same frame */
	printf("re-sending: code:%d length:%d addr:0x%x\r",bl_frame_code,bl_frame_count,bl_frame_addr);
	bl_send_frame(bl_current_frame,bl_current_framelen);
	bl_client_state = bl_state_waitack;
	break;
    }
  }
  
  fprintf(stderr,"\nTerminated\r\n");

#if 0
  /* Program Checking Phase */	
  
  while(1)
  {
  }
#endif

  kt_close_port();
  fclose(progfile);
}
